@@ -1,7 +1,7 @@
 /*
 	mpg123: main code of the program (not of the decoder...)
 
-	copyright 1995-2013 by the mpg123 project - free software under the terms of the LGPL 2.1
+	copyright 1995-2020 by the mpg123 project - free software under the terms of the LGPL 2.1
 	see COPYING and AUTHORS files in distribution or http://mpg123.org
 	initially written by Michael Hipp
 */
@@ -45,11 +45,11 @@
 #include "debug.h"
 
 static void usage(int err);
-static void want_usage(char* arg);
+static void want_usage(char* arg, topt *);
 static void long_usage(int err);
-static void want_long_usage(char* arg);
+static void want_long_usage(char* arg, topt *);
 static void print_title(FILE* o);
-static void give_version(char* arg);
+static void give_version(char* arg, topt *);
 
 struct parameter param = { 
   FALSE , /* aggressiv */
@@ -99,6 +99,11 @@ struct parameter param = {
 	,0 /* outscale */
 	,0 /* flags */
 	,0 /* force_rate */
+#ifndef NO_REAL
+	,2 /* fine resampling */
+#else
+	,0 /* NtoM resampler */
+#endif
 	,1 /* ICY */
 	,1024 /* resync_limit */
 	,0 /* smooth */
@@ -122,6 +127,7 @@ mpg123_handle *mh = NULL;
 off_t framenum;
 off_t frames_left;
 out123_handle *ao = NULL;
+
 static long output_propflags = 0;
 char *prgName = NULL;
 /* ThOr: pointers are not TRUE or FALSE */
@@ -130,10 +136,12 @@ struct httpdata htd;
 int fresh = TRUE;
 FILE* aux_out = NULL; /* Output for interesting information, normally on stdout to be parseable. */
 
+int stdout_is_term = FALSE; // It's an interactive terminal.
+int stderr_is_term = FALSE; // It's an interactive terminal.
+int got_played = 0; // State of attempted playback and success: 0, -1, 1
 int intflag = FALSE;
 int deathflag = FALSE;
 static int skip_tracks = 0;
-int OutputDescriptor;
 
 static int filept = -1;
 
@@ -168,23 +176,28 @@ static void catch_interrupt(void)
 {
 	intflag = TRUE;
 }
-static void handle_fatal_msg(const char *msg, size_t n)
+
+static void handle_fatal(void)
 {
-	if(msg && !param.quiet)
-		write(STDERR_FILENO, msg, n);
 	intflag = TRUE;
 	deathflag = TRUE;
 }
+
+static void handle_fatal_msg(const char *msg)
+{
+	if(msg && !param.quiet)
+		fprintf(stderr, "%s", msg);
+	handle_fatal();
+}
 static void catch_fatal_term(void)
 {
-	const char msg[] = "\nmpg123: death by SIGTERM\n";
-	handle_fatal_msg(msg, sizeof(msg));
+	handle_fatal_msg("\nmpg123: death by SIGTERM\n");
 }
 static void catch_fatal_pipe(void)
 {
 	/* If the SIGPIPE is because of piped stderr, trying to write
 	   in the signal handler hangs the program. */
-	handle_fatal_msg(NULL, 0);
+	handle_fatal();
 }
 #endif
 
@@ -225,7 +238,7 @@ static void play_prebuffer(void)
 	/* Ensure that the prebuffer bit has been posted. */
 	if(prebuffer_fill)
 	{
-		if(out123_play(ao, prebuffer, prebuffer_fill) < prebuffer_fill)
+		if(audio_play(ao, prebuffer, prebuffer_fill) < prebuffer_fill)
 		{
 			error("Deep trouble! Cannot flush to my output anymore!");
 			safe_exit(133);
@@ -268,16 +281,18 @@ void safe_exit(int code)
 {
 	char *dummy, *dammy;
 
-	if(prebuffer)
-		free(prebuffer);
-
 	dump_close();
 	if(!code)
 		controlled_drain();
 	if(intflag)
 		out123_drop(ao);
+	audio_cleanup();
 	out123_del(ao);
 
+	// Free the memory after the output is not working on it anymore!
+	if(prebuffer)
+		free(prebuffer);
+
 	if(mh != NULL) mpg123_delete(mh);
 
 	if(cleanup_mpg123) mpg123_exit();
@@ -304,14 +319,16 @@ static void check_fatal_output(int code)
 {
 	if(code)
 	{
-		if(!param.quiet)
+		if(out123_errcode(ao))
 			error2( "out123 error %i: %s"
 			,	out123_errcode(ao), out123_strerror(ao) );
+		else
+			error("fatal output (setup) error");
 		safe_exit(code);
 	}
 }
 
-static void set_output_module( char *arg )
+static void set_output_module(char *arg, topt *opts)
 {
 	unsigned int i;
 		
@@ -319,7 +336,7 @@ static void set_output_module( char *arg )
 	for(i=0; i< strlen( arg ); i++) {
 		if (arg[i] == ':') {
 			arg[i] = 0;
-			param.output_device = &arg[i+1];
+			getlopt_set_char(opts, "audiodevice", &arg[i+1]);
 			debug1("Setting output device: %s", param.output_device);
 			break;
 		}	
@@ -335,116 +352,125 @@ static void set_output_flag(int flag)
   else param.output_flags |= flag;
 }
 
-static void set_output_h(char *a)
+static void set_output_h(char *a, topt *opts)
 {
 	set_output_flag(OUT123_HEADPHONES);
 }
 
-static void set_output_s(char *a)
+static void set_output_s(char *a, topt *opts)
 {
 	set_output_flag(OUT123_INTERNAL_SPEAKER);
 }
 
-static void set_output_l(char *a)
+static void set_output_l(char *a, topt *opts)
 {
 	set_output_flag(OUT123_LINE_OUT);
 }
 
-static void set_output(char *arg)
+static void set_output(char *arg, topt *opts)
 {
 	/* If single letter, it's the legacy output switch for AIX/HP/Sun.
 	   If longer, it's module[:device] . If zero length, it's rubbish. */
 	if(strlen(arg) <= 1) switch(arg[0])
 	{
-		case 'h': set_output_h(arg); break;
-		case 's': set_output_s(arg); break;
-		case 'l': set_output_l(arg); break;
+		case 'h': set_output_h(arg, opts); break;
+		case 's': set_output_s(arg, opts); break;
+		case 'l': set_output_l(arg, opts); break;
 		default:
 			error1("\"%s\" is no valid output", arg);
 			safe_exit(1);
 	}
-	else set_output_module(arg);
+	else set_output_module(arg, opts);
 }
 
-static void set_verbose (char *arg)
+static void set_resample(char *arg, topt *opts)
+{
+	if(!strcasecmp("ntom", arg))
+		param.resample = 0;
+	else if(!strcasecmp("dirty", arg))
+		param.resample = 1;
+	else if(!strcasecmp("fine", arg))
+		param.resample = 2;
+	else
+	{
+		merror("\"%s\" is no valid resampler choice", arg);
+		safe_exit(1);
+	}
+}
+
+static void set_verbose (char *arg, topt *opts)
 {
     param.verbose++;
 }
 
-static void set_quiet (char *arg)
+static void set_quiet (char *arg, topt *opts)
 {
 	param.verbose=0;
 	param.quiet=TRUE;
 }
 
-static void set_out_wav(char *arg)
+static void set_out_wav(char *arg, topt *opts)
 {
 	param.output_module = "wav";
-	param.output_device = arg;
+	getlopt_set_char(opts, "audiodevice", arg);
 }
 
-void set_out_cdr(char *arg)
+void set_out_cdr(char *arg, topt *opts)
 {
 	param.output_module = "cdr";
-	param.output_device = arg;
+	getlopt_set_char(opts, "audiodevice", arg);
 }
 
-void set_out_au(char *arg)
+void set_out_au(char *arg, topt *opts)
 {
 	param.output_module = "au";
-	param.output_device = arg;
+	getlopt_set_char(opts, "audiodevice", arg);
 }
 
-void set_out_test(char *arg)
+void set_out_test(char *arg, topt *opts)
 {
 	param.output_module = "test";
-	param.output_device = NULL;
+	getlopt_set_char(opts, "audiodevice", NULL);
 }
 
-static void set_out_file(char *arg)
+static void set_out_file(char *arg, topt *opts)
 {
 	param.output_module = "raw";
-	param.output_device = arg;
+	getlopt_set_char(opts, "audiodevice", arg);
 }
 
-static void set_out_stdout(char *arg)
+static void set_out_stdout(char *arg, topt *opts)
 {
 	param.output_module = "raw";
-	param.output_device = NULL;
-}
-
-static void set_out_stdout1(char *arg)
-{
-	param.output_module = "raw";
-	param.output_device = NULL;
+	getlopt_set_char(opts, "audiodevice", NULL);
 }
 
 #if !defined (HAVE_SCHED_SETSCHEDULER) && !defined (HAVE_WINDOWS_H)
-static void realtime_not_compiled(char *arg)
+static void realtime_not_compiled(char *arg, topt *opts)
 {
 	fprintf(stderr,"Option '-T / --realtime' not compiled into this binary.\n");
 }
 #endif
 
-static int frameflag; /* ugly, but that's the way without hacking getlopt */
-static void set_frameflag(char *arg)
+static long frameflag; /* ugly, but that's the way without hacking getlopt */
+static void set_frameflag(char *arg, topt *opts)
 {
 	/* Only one mono flag at a time! */
 	if(frameflag & MPG123_FORCE_MONO) param.flags &= ~MPG123_FORCE_MONO;
 	param.flags |= frameflag;
 }
-static void unset_frameflag(char *arg)
+static void unset_frameflag(char *arg, topt *opts)
 {
 	param.flags &= ~frameflag;
 }
 
-static int appflag; /* still ugly, but works */
-static void set_appflag(char *arg)
+long appflag; /* still ugly, but works */
+static void set_appflag(char *arg, topt *opts)
 {
 	param.appflags |= appflag;
 }
 
-static void list_output_modules(char *arg)
+static void list_output_modules(char *arg, topt *opts)
 {
 	char **names = NULL;
 	char **descr = NULL;
@@ -503,27 +529,28 @@ topt opts[] = {
 	{'4', "4to1",        GLO_INT,  0, &param.down_sample, 2},
 	{'t', "test",        GLO_INT,  set_out_test, NULL, 0},
 	{'s', "stdout",      GLO_INT,  set_out_stdout,  NULL, 0},
-	{'S', "STDOUT",      GLO_INT,  set_out_stdout1, NULL, 0},
 	{'O', "outfile",     GLO_ARG | GLO_CHAR, set_out_file, NULL, 0},
 	{'c', "check",       GLO_INT,  0, &param.checkrange, TRUE},
 	{'v', "verbose",     0,        set_verbose, 0,           0},
 	{'q', "quiet",       0,        set_quiet,   0,           0},
-	{'y', "no-resync",      GLO_INT,  set_frameflag, &frameflag, MPG123_NO_RESYNC},
+	{'y', "no-resync",       GLO_LONG,  set_frameflag, &frameflag, MPG123_NO_RESYNC},
+	{'F', "no-frankenstein", GLO_LONG,  set_frameflag, &frameflag, MPG123_NO_FRANKENSTEIN},
 	/* compatibility, no-resync is to be used nowadays */
-	{0, "resync",      GLO_INT,  set_frameflag, &frameflag, MPG123_NO_RESYNC},
-	{'0', "single0",     GLO_INT,  set_frameflag, &frameflag, MPG123_MONO_LEFT},
-	{0,   "left",        GLO_INT,  set_frameflag, &frameflag, MPG123_MONO_LEFT},
-	{'1', "single1",     GLO_INT,  set_frameflag, &frameflag, MPG123_MONO_RIGHT},
-	{0,   "right",       GLO_INT,  set_frameflag, &frameflag, MPG123_MONO_RIGHT},
-	{'m', "singlemix",   GLO_INT,  set_frameflag, &frameflag, MPG123_MONO_MIX},
-	{0,   "mix",         GLO_INT,  set_frameflag, &frameflag, MPG123_MONO_MIX},
-	{0,   "mono",        GLO_INT,  set_frameflag, &frameflag, MPG123_MONO_MIX},
-	{0,   "stereo",      GLO_INT,  set_frameflag, &frameflag, MPG123_FORCE_STEREO},
+	{0, "resync",        GLO_LONG,  set_frameflag, &frameflag, MPG123_NO_RESYNC},
+	{'0', "single0",     GLO_LONG,  set_frameflag, &frameflag, MPG123_MONO_LEFT},
+	{0,   "left",        GLO_LONG,  set_frameflag, &frameflag, MPG123_MONO_LEFT},
+	{'1', "single1",     GLO_LONG,  set_frameflag, &frameflag, MPG123_MONO_RIGHT},
+	{0,   "right",       GLO_LONG,  set_frameflag, &frameflag, MPG123_MONO_RIGHT},
+	{'m', "singlemix",   GLO_LONG,  set_frameflag, &frameflag, MPG123_MONO_MIX},
+	{0,   "mix",         GLO_LONG,  set_frameflag, &frameflag, MPG123_MONO_MIX},
+	{0,   "mono",        GLO_LONG,  set_frameflag, &frameflag, MPG123_MONO_MIX},
+	{0,   "stereo",      GLO_LONG,  set_frameflag, &frameflag, MPG123_FORCE_STEREO},
 	{0,   "reopen",      GLO_INT,  0, &param.force_reopen, 1},
 	{'g', "gain",        GLO_ARG | GLO_LONG, 0, &param.gain,    0},
 	{'r', "rate",        GLO_ARG | GLO_LONG, 0, &param.force_rate,  0},
-	{0,   "8bit",        GLO_INT,  set_frameflag, &frameflag, MPG123_FORCE_8BIT},
-	{0,   "float",       GLO_INT,  set_frameflag, &frameflag, MPG123_FORCE_FLOAT},
+	{0, "resample",      GLO_ARG | GLO_CHAR, set_resample, NULL,  0},
+	{0,   "8bit",        GLO_LONG,  set_frameflag, &frameflag, MPG123_FORCE_8BIT},
+	{0,   "float",       GLO_LONG,  set_frameflag, &frameflag, MPG123_FORCE_FLOAT},
 	{0,   "headphones",  0,                  set_output_h, 0,0},
 	{0,   "speaker",     0,                  set_output_s, 0,0},
 	{0,   "lineout",     0,                  set_output_l, 0,0},
@@ -533,6 +560,7 @@ topt opts[] = {
 	{'f', "scale",       GLO_ARG | GLO_LONG, 0, &param.outscale,   0},
 	{'n', "frames",      GLO_ARG | GLO_LONG, 0, &param.frame_number,  0},
 #ifdef HAVE_TERMIOS
+	{0, "no-visual",     GLO_INT,  0, &param.term_visual, FALSE},
 	{'C', "control",     GLO_INT,  0, &param.term_ctrl, TRUE},
 	{0, "no-control",    GLO_INT,  0, &param.term_ctrl, FALSE},
 	{0,   "ctrlusr1",    GLO_ARG | GLO_CHAR, 0, &param.term_usr1, 0},
@@ -584,14 +612,14 @@ topt opts[] = {
 	{'w', "wav",         GLO_ARG | GLO_CHAR, set_out_wav, 0, 0 },
 	{0, "cdr",           GLO_ARG | GLO_CHAR, set_out_cdr, 0, 0 },
 	{0, "au",            GLO_ARG | GLO_CHAR, set_out_au, 0, 0 },
-	{0,   "gapless",	 GLO_INT,  set_frameflag, &frameflag, MPG123_GAPLESS},
-	{0,   "no-gapless", GLO_INT, unset_frameflag, &frameflag, MPG123_GAPLESS},
-	{0, "no-infoframe", GLO_INT, set_frameflag, &frameflag, MPG123_IGNORE_INFOFRAME},
+	{0,   "gapless",	   GLO_LONG, set_frameflag,   &frameflag, MPG123_GAPLESS},
+	{0,   "no-gapless",  GLO_LONG, unset_frameflag, &frameflag, MPG123_GAPLESS},
+	{0, "no-infoframe",  GLO_LONG, set_frameflag,   &frameflag, MPG123_IGNORE_INFOFRAME},
 	{'?', "help",            0,  want_usage, 0,           0 },
 	{0 , "longhelp" ,        0,  want_long_usage, 0,      0 },
 	{0 , "version" ,         0,  give_version, 0,         0 },
 	{'l', "listentry",       GLO_ARG | GLO_LONG, 0, &param.listentry, 0 },
-	{0, "continue", GLO_INT, set_appflag, &appflag, MPG123APP_CONTINUE },
+	{0, "continue",      GLO_LONG, set_appflag, &appflag, MPG123APP_CONTINUE },
 	{0, "rva-mix",         GLO_INT,  0, &param.rva, 1 },
 	{0, "rva-radio",         GLO_INT,  0, &param.rva, 1 },
 	{0, "rva-album",         GLO_INT,  0, &param.rva, 2 },
@@ -608,20 +636,20 @@ topt opts[] = {
 	{0, "resync-limit", GLO_ARG | GLO_LONG, 0, &param.resync_limit, 0},
 	{0, "pitch", GLO_ARG|GLO_DOUBLE, 0, &param.pitch, 0},
 #ifdef NETWORK
-	{0, "ignore-mime", GLO_INT, set_appflag, &appflag, MPG123APP_IGNORE_MIME },
+	{0, "ignore-mime", GLO_LONG, set_appflag, &appflag, MPG123APP_IGNORE_MIME },
 #endif
-	{0, "lyrics", GLO_INT, set_appflag, &appflag, MPG123APP_LYRICS},
+	{0, "lyrics", GLO_LONG, set_appflag, &appflag, MPG123APP_LYRICS},
 	{0, "keep-open", GLO_INT, 0, &param.keep_open, 1},
 	{0, "utf8", GLO_INT, 0, &param.force_utf8, 1},
-	{0, "fuzzy", GLO_INT,  set_frameflag, &frameflag, MPG123_FUZZY},
+	{0, "fuzzy", GLO_LONG,  set_frameflag, &frameflag, MPG123_FUZZY},
 	{0, "index-size", GLO_ARG|GLO_LONG, 0, &param.index_size, 0},
-	{0, "no-seekbuffer", GLO_INT, unset_frameflag, &frameflag, MPG123_SEEKBUFFER},
+	{0, "no-seekbuffer", GLO_LONG, unset_frameflag, &frameflag, MPG123_SEEKBUFFER},
 	{'e', "encoding", GLO_ARG|GLO_CHAR, 0, &param.force_encoding, 0},
 	{0, "preframes", GLO_ARG|GLO_LONG, 0, &param.preframes, 0},
-	{0, "skip-id3v2", GLO_INT, set_frameflag, &frameflag, MPG123_SKIP_ID3V2},
+	{0, "skip-id3v2", GLO_LONG, set_frameflag, &frameflag, MPG123_SKIP_ID3V2},
 	{0, "streamdump", GLO_ARG|GLO_CHAR, 0, &param.streamdump, 0},
 	{0, "icy-interval", GLO_ARG|GLO_LONG, 0, &param.icy_interval, 0},
-	{0, "ignore-streamlength", GLO_INT, set_frameflag, &frameflag, MPG123_IGNORE_STREAMLENGTH},
+	{0, "ignore-streamlength", GLO_LONG, set_frameflag, &frameflag, MPG123_IGNORE_STREAMLENGTH},
 	{0, "name", GLO_ARG|GLO_CHAR, 0, &param.name, 0},
 	{0, "devbuffer", GLO_ARG|GLO_DOUBLE, 0, &param.device_buffer, 0},
 	{0, 0, 0, 0, 0, 0}
@@ -648,8 +676,10 @@ int open_track(char *fname)
 	httpdata_reset(&htd);
 	if(MPG123_OK != mpg123_param(mh, MPG123_ICY_INTERVAL, 0, 0))
 	error1("Cannot (re)set ICY interval: %s", mpg123_strerror(mh));
+	mpg123_param(mh, MPG123_REMOVE_FLAGS, MPG123_NO_PEEK_END, 0);
 	if(!strcmp(fname, "-"))
 	{
+		mpg123_param(mh, MPG123_ADD_FLAGS, MPG123_NO_PEEK_END, 0);
 		filept = STDIN_FILENO;
 #ifdef WIN32
 		_setmode(STDIN_FILENO, _O_BINARY);
@@ -792,7 +822,7 @@ int play_frame(void)
 				bytes -= missing;
 				prebuffer_fill += missing;
 			}
-			if(   out123_play(ao, prebuffer, prebuffer_fill) < prebuffer_fill
+			if(   audio_play(ao, prebuffer, prebuffer_fill) < prebuffer_fill
 			   && !intflag )
 			{
 				error("Deep trouble! Cannot flush to my output anymore!");
@@ -804,7 +834,7 @@ int play_frame(void)
 		   I wonder if that makes us miss errors. Actual issues should
 		   just be postponed. */
 		if(bytes && !intflag) /* Previous piece could already be interrupted. */
-		if(out123_play(ao, playbuf, bytes) < bytes && !intflag)
+		if(audio_play(ao, playbuf, bytes) < bytes && !intflag)
 		{
 			error("Deep trouble! Cannot flush to my output anymore!");
 			safe_exit(133);
@@ -813,14 +843,20 @@ int play_frame(void)
 	/* Special actions and errors. */
 	if(mc != MPG123_OK)
 	{
-		if(mc == MPG123_ERR || mc == MPG123_DONE)
+		if(mc == MPG123_ERR || mc == MPG123_DONE || mc == MPG123_NEED_MORE)
 		{
-			if(mc == MPG123_ERR) error1("...in decoding next frame: %s", mpg123_strerror(mh));
+			if(!param.quiet && mc == MPG123_ERR) error1("...in decoding next frame: %s", mpg123_strerror(mh));
+			if(!param.quiet && mc == MPG123_NEED_MORE)
+				error("The decoder expected more. Old libmpg123 that does this for non-feed readers?");
+			if(mc != MPG123_DONE && !got_played)
+				got_played = -1;
 			return 0;
 		}
 		if(mc == MPG123_NO_SPACE)
 		{
 			error("I have not enough output space? I didn't plan for this.");
+			if(!got_played)
+				got_played = -1;
 			return 0;
 		}
 		if(mc == MPG123_NEW_FORMAT)
@@ -835,7 +871,9 @@ int play_frame(void)
 				minbytes = out123_encsize(encoding)*channels*384;
 			else
 				minbytes = 0;
-			if(param.verbose > 2)
+			if(framenum && param.verbose)
+				print_stat(mh,0,ao,0,&param);
+			if(param.verbose > 1)
 			{
 				const char* encname = out123_enc_name(encoding);
 				fprintf( stderr
@@ -843,20 +881,24 @@ int play_frame(void)
 				,	rate, channels, encname ? encname : "???" );
 			}
 			new_header = TRUE;
-			check_fatal_output(out123_start(ao, rate, channels, encoding));
+			
+			check_fatal_output(audio_prepare(ao, mh, rate, channels, encoding));
 			/* We may take some time feeding proper data, so pause by default. */
 			out123_pause(ao);
 		}
 	}
-	if(new_header && !param.quiet)
+	if((param.verbose > 3 || new_header) && !param.quiet)
 	{
 		new_header = FALSE;
+		if(framenum)
+			print_stat(mh,0,ao,0,&param);
 		fprintf(stderr, "\n");
 		if(param.verbose > 1)
 			print_header(mh);
 		else
 			print_header_compact(mh);
 	}
+
 	return 1;
 }
 
@@ -922,6 +964,8 @@ int main(int sys_argc, char ** sys_argv)
 	char *fname;
 	int libpar = 0;
 	mpg123_pars *mp;
+	int args_utf8 = FALSE;
+	int pl_utf8   = FALSE;
 #if !defined(WIN32) && !defined(GENERIC)
 	struct timeval start_time;
 #endif
@@ -932,6 +976,7 @@ int main(int sys_argc, char ** sys_argv)
 		error("Cannot convert command line to UTF8!");
 		safe_exit(76);
 	}
+	args_utf8 = TRUE;
 #else
 	argv = sys_argv;
 	argc = sys_argc;
@@ -1014,19 +1059,25 @@ int main(int sys_argc, char ** sys_argv)
 	/* Detect terminal on input side, enable control by default. */
 	param.term_ctrl = !(term_width(STDIN_FILENO) < 0);
 #endif
+	stderr_is_term = term_width(STDERR_FILENO) >= 0;
+	stdout_is_term = term_width(STDOUT_FILENO) >= 0;
 	while ((result = getlopt(argc, argv, opts)))
 	switch (result) {
 		case GLO_UNKNOWN:
-			fprintf (stderr, "%s: Unknown option \"%s\".\n", 
-				prgName, loptarg);
+			print_outstr(stderr, prgName, 0, stderr_is_term);
+			fprintf(stderr, ": Unknown option \"");
+			print_outstr(stderr, loptarg, 0, stderr_is_term);
+			fprintf(stderr, "\".\n");
 			usage(1);
 		case GLO_NOARG:
-			fprintf (stderr, "%s: Missing argument for option \"%s\".\n",
-				prgName, loptarg);
+			print_outstr(stderr, prgName, 0, stderr_is_term);
+			fprintf(stderr, ": Missing argument for option \"%s\".\n", loptarg);
 			usage(1);
 	}
 	/* Do this _after_ parameter parsing. */
+	utf8force = param.force_utf8;
 	check_locale(); /* Check/set locale; store if it uses UTF-8. */
+	meta_show_lyrics = APPFLAG(MPG123APP_LYRICS);
 
 	if(param.list_cpu)
 	{
@@ -1086,8 +1137,6 @@ int main(int sys_argc, char ** sys_argv)
 	    && ++libpar
 	    && MPG123_OK == (result = mpg123_par(mp, MPG123_RVA, param.rva, 0))
 	    && ++libpar
-	    && MPG123_OK == (result = mpg123_par(mp, MPG123_FORCE_RATE, param.force_rate, 0))
-	    && ++libpar
 	    && MPG123_OK == (result = mpg123_par(mp, MPG123_DOWNSPEED, param.halfspeed, 0))
 	    && ++libpar
 	    && MPG123_OK == (result = mpg123_par(mp, MPG123_UPSPEED, param.doublespeed, 0))
@@ -1172,8 +1221,7 @@ int main(int sys_argc, char ** sys_argv)
 	ao = out123_new();
 	if(!ao)
 	{
-		if(!param.quiet)
-			error("Failed to allocate output.");
+		error("Failed to allocate output.");
 		safe_exit(97);
 	}
 	if
@@ -1187,8 +1235,7 @@ int main(int sys_argc, char ** sys_argv)
 	|| out123_param_float(ao, OUT123_DEVICEBUFFER, param.device_buffer)
 	)
 	{
-		if(!param.quiet)
-			error("Error setting output parameters. Do you need a usage reminder?");
+		error("Error setting output parameters. Do you need a usage reminder?");
 		safe_exit(98);
 	}
 	check_fatal_output(out123_set_buffer(ao, param.usebuffer*1024));
@@ -1196,7 +1243,8 @@ int main(int sys_argc, char ** sys_argv)
 	,	param.output_module, param.output_device ));
 	out123_getparam_int(ao, OUT123_PROPFLAGS, &output_propflags);
 
-	if(!param.remote) prepare_playlist(argc, argv);
+	if(!param.remote)
+		prepare_playlist(argc, argv, args_utf8, &pl_utf8);
 
 #if !defined(WIN32) && !defined(GENERIC)
 	/* Remote mode is special... but normal console and terminal-controlled operation needs to catch the SIGINT.
@@ -1209,7 +1257,11 @@ int main(int sys_argc, char ** sys_argv)
 	catchsignal(SIGPIPE, catch_fatal_pipe);
 #endif
 	/* Now either check caps myself or query buffer for that. */
-	audio_capabilities(ao, mh);
+	if(audio_setup(ao, mh))
+	{
+		error("Failed to set up audio output.");
+		safe_exit(7);
+	}
 
 	if(param.remote) {
 		int ret;
@@ -1246,8 +1298,14 @@ int main(int sys_argc, char ** sys_argv)
 		if(!APPFLAG(MPG123APP_CONTINUE)) frames_left = param.frame_number;
 
 		debug1("Going to play %s", strcmp(fname, "-") ? fname : "standard input");
-
-		if(intflag || !open_track(fname))
+		// If a previous track did not cause error, we forget the success to
+		// catch an error in this track.
+		if(got_played > -1)
+			got_played = 0;
+		int open_track_ret = open_track(fname);
+		if(!open_track_ret)
+			got_played = -1;
+		if(intflag || !open_track_ret)
 		{
 #ifdef HAVE_TERMIOS
 			/* We need the opportunity to cancel in case of --loop -1 . */
@@ -1301,7 +1359,12 @@ int main(int sys_argc, char ** sys_argv)
 			size_t plfill;
 			long plloop;
 			plpos = playlist_pos(&plfill, &plloop);
-			if(newdir) fprintf(stderr, "Directory: %s\n", dirname);
+			if(newdir)
+			{
+				fprintf(stderr, "Directory: ");
+				print_outstr(stderr, dirname, 0, stderr_is_term);
+				fprintf(stderr, "\n");
+			}
 
 #ifdef HAVE_TERMIOS
 		/* Reminder about terminal usage. */
@@ -1316,8 +1379,10 @@ int main(int sys_argc, char ** sys_argv)
 					fprintf( stderr, "Repeating %ld %s.\n"
 					,	plloop-1, plloop > 2 ? "times" : "time" );
 			}
-			fprintf( stderr, "Playing MPEG stream %"SIZE_P" of %"SIZE_P": %s ...\n"
-			,	(size_p)plpos, (size_p)plfill, filename );
+			fprintf( stderr, "Playing MPEG stream %"SIZE_P" of %"SIZE_P": "
+			,	(size_p)plpos, (size_p)plfill );
+			print_outstr(stderr, filename, 0, stderr_is_term);
+			fprintf(stderr, " ...\n");
 			if(htd.icy_name.fill) fprintf(stderr, "ICY-NAME: %s\n", htd.icy_name.p);
 			if(htd.icy_url.fill)  fprintf(stderr, "ICY-URL: %s\n",  htd.icy_url.p);
 		}
@@ -1327,14 +1392,20 @@ int main(int sys_argc, char ** sys_argv)
 	term_type = getenv("TERM");
 	if(term_type && param.xterm_title)
 	{
+		mpg123_string titlestring;
+		mpg123_init_string(&titlestring);
+		outstr(&titlestring, filename, pl_utf8, 1);
+		const char *ts = titlestring.fill ? titlestring.p : "???";
+
 		if(!strncmp(term_type,"xterm",5) || !strncmp(term_type,"rxvt",4))
-		fprintf(stderr, "\033]0;%s\007", filename);
+			fprintf(stderr, "\033]0;%s\007", ts);
 		else if(!strncmp(term_type,"screen",6))
-		fprintf(stderr, "\033k%s\033\\", filename);
+			fprintf(stderr, "\033k%s\033\\", ts);
 		else if(!strncmp(term_type,"iris-ansi",9))
-		fprintf(stderr, "\033P1.y %s\033\\\033P3.y%s\033\\", filename, filename);
+			fprintf(stderr, "\033P1.y %s\033\\\033P3.y%s\033\\", ts, ts);
 
 		fflush(stderr); /* Paranoia: will the buffer buffer the escapes? */
+		mpg123_free_string(&titlestring);
 	}
 }
 #endif
@@ -1361,12 +1432,19 @@ int main(int sys_argc, char ** sys_argv)
 				}
 			}
 			if(!play_frame()) break;
+			// At least one successful frame makes the whole run successful, unless
+			// a previous run erred.
+			if(!got_played)
+				got_played = 1;
 			if(!param.quiet)
 			{
 				meta = mpg123_meta_check(mh);
 				if(meta & (MPG123_NEW_ID3|MPG123_NEW_ICY))
 				{
-					if(meta & MPG123_NEW_ID3) print_id3_tag(mh, param.long_id3, stderr);
+					if(framenum && param.verbose)
+						print_stat(mh,0,ao,0,&param);
+					if(meta & MPG123_NEW_ID3) print_id3_tag( mh, param.long_id3
+					,	stderr, term_width(STDERR_FILENO) );
 					if(meta & MPG123_NEW_ICY) print_icy(mh, stderr);
 
 #ifdef HAVE_TERMIOS
@@ -1377,7 +1455,7 @@ int main(int sys_argc, char ** sys_argv)
 			}
 			if(!fresh && param.verbose)
 			{
-				if(param.verbose > 1 || !(framenum & 0x7)) print_stat(mh,0,ao,1);
+				if(param.verbose > 1 || !(framenum & 0x7)) print_stat(mh,0,ao,1,&param);
 			}
 #ifdef HAVE_TERMIOS
 			if(!param.term_ctrl) continue;
@@ -1387,7 +1465,7 @@ int main(int sys_argc, char ** sys_argv)
 
 	if(!param.smooth && !intflag)
 		controlled_drain();
-	if(param.verbose) print_stat(mh,0,ao,0);
+	if(param.verbose) print_stat(mh,0,ao,0,&param);
 
 	if(!param.quiet)
 	{
@@ -1425,8 +1503,8 @@ int main(int sys_argc, char ** sys_argv)
 	/* Free up memory used by playlist */    
 	if(!param.remote) free_playlist();
 
-	safe_exit(0); /* That closes output and restores terminal, too. */
-	return 0;
+	safe_exit(got_played < 0 ? 1 : 0); /* That closes output and restores terminal, too. */
+	return -1; // Should never be reached.
 }
 
 static void print_title(FILE *o)
@@ -1480,7 +1558,7 @@ static void usage(int err)  /* print syntax & exit */
 	safe_exit(err);
 }
 
-static void want_usage(char* arg)
+static void want_usage(char* arg, topt *opts)
 {
 	usage(0);
 }
@@ -1504,6 +1582,7 @@ static void long_usage(int err)
 	fprintf(o," -n     --frames <n>       play only <n> frames of every stream\n");
 	fprintf(o,"        --fuzzy            Enable fuzzy seeks (guessing byte offsets or using approximate seek points from Xing TOC)\n");
 	fprintf(o," -y     --no-resync        DISABLES resync on error (--resync is deprecated)\n");
+	fprintf(o," -F     --no-frankenstein  disable support for Frankenstein streams\n");
 #ifdef NETWORK
 	fprintf(o," -p <f> --proxy <f>        set WWW proxy\n");
 	fprintf(o," -u     --auth             set auth values for HTTP access\n");
@@ -1531,7 +1610,6 @@ static void long_usage(int err)
 	fprintf(o,"        --list-modules     list the available modules\n");
 	fprintf(o," -a <d> --audiodevice <d>  select audio device (depending on chosen module)\n");
 	fprintf(o," -s     --stdout           write raw audio to stdout (host native format)\n");
-	fprintf(o," -S     --STDOUT           play AND output stream (not implemented yet)\n");
 	fprintf(o," -w <f> --wav <f>          write samples as WAV file in <f> (- is stdout)\n");
 	fprintf(o,"        --au <f>           write samples as Sun AU file in <f> (- is stdout)\n");
 	fprintf(o,"        --cdr <f>          write samples as raw CD audio file in <f> (- is stdout)\n");
@@ -1557,16 +1635,18 @@ static void long_usage(int err)
 	fprintf(o," -m     --mono --mix       mix stereo to mono\n");
 	fprintf(o,"        --stereo           duplicate mono channel\n");
 	fprintf(o," -r     --rate             force a specific audio output rate\n");
+	fprintf(o,"        --resample <s>     choose resampling mode for forced rate:\n"
+	          "                           NtoM, dirty, fine (default)\n");
 	fprintf(o," -2     --2to1             2:1 downsampling\n");
 	fprintf(o," -4     --4to1             4:1 downsampling\n");
-  fprintf(o,"        --pitch <value>    set hardware pitch (speedup/down, 0 is neutral; 0.05 is 5%%)\n");
+	fprintf(o,"        --pitch <value>    set pitch (speedup/down, 0 is neutral; 0.05 is 5%%)\n");
 	fprintf(o,"        --8bit             force 8 bit output\n");
 	fprintf(o,"        --float            force floating point output (internal precision)\n");
 	fprintf(o," -e <c> --encoding <c>     force a specific encoding (%s)\n"
 	,	enclist != NULL ? enclist->p : "OOM!");
 	fprintf(o," -d n   --doublespeed n    play only every nth frame\n");
 	fprintf(o," -h n   --halfspeed   n    play every frame n times\n");
-	fprintf(o,"        --equalizer        exp.: scales freq. bands acrd. to 'equalizer.dat'\n");
+	fprintf(o,"        --equalizer   f    scales freq. bands acrd. to given file\n");
 	fprintf(o,"        --gapless          remove padding/junk on mp3s (best with Lame tag)\n");
 	fprintf(o,"                           This is on by default when libmpg123 supports it.\n");
 	fprintf(o,"        --no-gapless       disable gapless mode, not remove padding/junk\n");
@@ -1590,6 +1670,8 @@ static void long_usage(int err)
 	#ifdef HAVE_TERMIOS
 	fprintf(o," -C     --control          enable terminal control keys (else auto detect)\n");
 	fprintf(o,"        --no-control       disable terminal control keys (disable auto detect)\n");
+	fprintf(o,"        --no-visual        disable visual enhancements in output (hide cursor,\n"
+	          "                           reverse video), alternative to TERM=dumb\n");
 	fprintf(o,"        --ctrlusr1 <c>     control key (characer) to map to SIGUSR1\n");
 	fprintf(o,"                           (default is for stop/start)\n");
 	fprintf(o,"        --ctrlusr2 <c>     control key (characer) to map to SIGUSR2\n");
@@ -1628,12 +1710,12 @@ static void long_usage(int err)
 	safe_exit(err);
 }
 
-static void want_long_usage(char* arg)
+static void want_long_usage(char* arg, topt *opts)
 {
 	long_usage(0);
 }
 
-static void give_version(char* arg)
+static void give_version(char* arg, topt *opts)
 {
 	fprintf(stdout, PACKAGE_NAME" "PACKAGE_VERSION"\n");
 	safe_exit(0);
